import asyncio
import json
import random
from pathlib import Path

import pyperclip
from ollama import Message, ResponseError
from textual import on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.events import Click
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import (
    LoadingIndicator,
    Markdown,
    Static,
    TabbedContent,
)

from oterm.app.chat_edit import ChatEdit
from oterm.app.chat_rename import ChatRename
from oterm.app.prompt_history import PromptHistory
from oterm.app.widgets.image import ImageAdded
from oterm.app.widgets.prompt import FlexibleInput
from oterm.ollamaclient import OllamaLLM, Options
from oterm.store.store import Store
from oterm.tools import available as available_tool_defs
from oterm.types import Author, Tool
from oterm.utils import parse_response


class ChatContainer(Widget):
    ollama = OllamaLLM()
    messages: reactive[list[tuple[int, Author, str, list[str]]]] = reactive([])
    chat_name: str
    system: str | None
    format: str
    parameters: Options
    keep_alive: int = 5
    images: list[tuple[Path, str]] = []
    tools: list[Tool] = []
    BINDINGS = [
        Binding("up", "history", "history"),
        Binding(
            "escape", "cancel_inference", "cancel inference", show=False, priority=True
        ),
    ]

    def __init__(
        self,
        *children: Widget,
        db_id: int,
        chat_name: str,
        model: str = "llama3.2",
        messages: list[tuple[int, Author, str, list[str]]] = [],
        system: str | None = None,
        format: str = "",
        parameters: Options,
        keep_alive: int = 5,
        tools: list[Tool] = [],
        **kwargs,
    ) -> None:
        super().__init__(*children, **kwargs)
        history = []
        # This is wrong, the images should be a list of Image objects
        # See https://github.com/ollama/ollama-python/issues/375
        # Temp fix is to do msg.images = images  # type: ignore
        for _, author, message, images in messages:
            msg = Message(
                role="user" if author == Author.USER else "assistant",
                content=(
                    message
                    if author == Author.USER
                    else parse_response(message)["response"]
                ),
            )
            msg.images = images  # type: ignore
            history.append(msg)
        used_tool_defs = [
            tool_def for tool_def in available_tool_defs if tool_def["tool"] in tools
        ]

        self.ollama = OllamaLLM(
            model=model,
            system=system,
            format=format,
            options=parameters,
            keep_alive=keep_alive,
            history=history,  # type: ignore
            tool_defs=used_tool_defs,
        )

        self.chat_name = chat_name
        self.db_id = db_id
        self.messages = messages
        self.system = system
        self.format = format
        self.parameters = parameters
        self.keep_alive = keep_alive
        self.tools = tools
        self.loaded = False
        self.loading = False
        self.images = []

    def on_mount(self) -> None:
        self.query_one("#prompt").focus()

    async def load_messages(self) -> None:
        if self.loaded or self.loading:
            return
        self.loading = True
        message_container = self.query_one("#messageContainer")
        for _, author, message, images in self.messages:
            chat_item = ChatItem()
            chat_item.text = (
                message
                if author == Author.USER
                else parse_response(message)["formatted_output"]
            )
            chat_item.author = author
            await message_container.mount(chat_item)
        message_container.scroll_end()
        self.loading = False
        self.loaded = True

    @on(FlexibleInput.Submitted)
    async def on_submit(self, event: FlexibleInput.Submitted) -> None:
        message = event.value
        input = event.input
        message_container = self.query_one("#messageContainer")

        if not message.strip():
            input.clear()
            input.focus()
            return

        async def response_task() -> None:
            input.clear()
            user_chat_item = ChatItem()
            user_chat_item.text = message
            user_chat_item.author = Author.USER
            message_container.mount(user_chat_item)

            response_chat_item = ChatItem()
            response_chat_item.author = Author.OLLAMA
            message_container.mount(response_chat_item)
            loading = LoadingIndicator()
            await message_container.mount(loading)
            message_container.scroll_end()

            try:
                response = ""

                # Ollama does not support streaming with tools, so we need to use completion
                if self.tools:
                    response = await self.ollama.completion(
                        prompt=message, images=[img for _, img in self.images]
                    )
                    response_chat_item.text = response

                else:
                    async for text in self.ollama.stream(
                        message, [img for _, img in self.images]
                    ):
                        response = text
                        response_chat_item.text = text

                # Parse the response for special tags
                parsed = parse_response(response)

                # To not exhaust the tokens, remove the thought process from the history (it seems to be the common practice)
                self.ollama.history[-1].content = parsed["response"]

                # Update the text according to the new formatted response.
                response_chat_item.text = parsed["formatted_output"]

                if message_container.can_view_partial(response_chat_item):
                    message_container.scroll_end()

                # Save to db
                store = await Store.get_store()
                id = await store.save_message(
                    id=None,
                    chat_id=self.db_id,
                    author=Author.USER.value,
                    text=message,
                    images=[img for _, img in self.images],
                )
                self.messages.append(
                    (id, Author.USER, message, [img for _, img in self.images])
                )

                id = await store.save_message(
                    id=None,
                    chat_id=self.db_id,
                    author=Author.OLLAMA.value,
                    text=response,
                )
                self.messages.append((id, Author.OLLAMA, response, []))
                self.images = []

            except asyncio.CancelledError:
                user_chat_item.remove()
                response_chat_item.remove()
                input.text = message
            except ResponseError as e:
                user_chat_item.remove()
                response_chat_item.remove()
                self.app.notify(
                    f"There was an error running your request: {e}", severity="error"
                )
                message_container.scroll_end()

            finally:
                loading.remove()
                input.focus()

        self.inference_task = asyncio.create_task(response_task())

    def key_escape(self) -> None:
        if hasattr(self, "inference_task"):
            self.inference_task.cancel()

    @work
    async def action_edit_chat(self) -> None:
        screen = ChatEdit(
            model=self.ollama.model,
            system=self.system or "",
            parameters=self.parameters,
            format=self.format,
            keep_alive=self.keep_alive,
            edit_mode=True,
            tools=self.tools,
        )

        model_info = await self.app.push_screen_wait(screen)
        if model_info is None:
            return
        model: dict = json.loads(model_info)
        self.system = model.get("system")
        self.format = model.get("format", "")
        self.keep_alive = model.get("keep_alive", 5)
        self.parameters = Options(**model.get("parameters", {}))
        self.tools = [Tool(**t) for t in model.get("tools", [])]
        store = await Store.get_store()
        await store.edit_chat(  # type: ignore
            id=self.db_id,
            name=self.chat_name,
            system=model["system"],
            format=model["format"],
            parameters=model["parameters"],
            keep_alive=model["keep_alive"],
            tools=model["tools"],
        )

        # load the history from messages
        history: list[Message] = []
        # This is wrong, the images should be a list of Image objects
        # See https://github.com/ollama/ollama-python/issues/375
        # Temp fix is to do msg.images = images  # type: ignore
        for _, author, message, images in self.messages:
            msg = Message(
                role="user" if author == Author.USER else "assistant",
                content=message,
            )
            msg.images = images  # type: ignore
            history.append(msg)

        used_tool_defs = [
            tool_def
            for tool_def in available_tool_defs
            if tool_def["tool"] in self.tools
        ]

        self.ollama = OllamaLLM(
            model=model["name"],
            system=model["system"],
            format=model["format"],
            options=self.parameters,
            keep_alive=model["keep_alive"],
            history=history,  # type: ignore
            tool_defs=used_tool_defs,
        )

    @work
    async def action_rename_chat(self) -> None:
        store = await Store.get_store()

        screen = ChatRename(self.chat_name)
        new_name = await self.app.push_screen_wait(screen)
        if new_name is None:
            return
        tabs = self.app.query_one(TabbedContent)
        await store.rename_chat(self.db_id, new_name)
        tabs.get_tab(f"chat-{self.db_id}").update(new_name)

    async def action_clear_chat(self) -> None:
        self.messages = []
        self.images = []
        self.ollama = OllamaLLM(
            model=self.ollama.model,
            system=self.ollama.system,
            format=self.ollama.format,  # type: ignore
            options=self.parameters,
            keep_alive=self.ollama.keep_alive,
            history=[],  # type: ignore
            tool_defs=self.ollama.tool_defs,
        )
        msg_container = self.query_one("#messageContainer")
        for child in msg_container.children:
            child.remove()
        store = await Store.get_store()
        await store.clear_chat(self.db_id)

    async def action_regenerate_llm_message(self) -> None:
        if not self.messages[-1:]:
            return

        # Remove last Ollama response from UI and regenerate it
        response_message_id = self.messages[-1][0]
        self.messages.pop()
        message_container = self.query_one("#messageContainer")
        message_container.children[-1].remove()
        response_chat_item = ChatItem()
        response_chat_item.author = Author.OLLAMA
        message_container.mount(response_chat_item)
        loading = LoadingIndicator()
        await message_container.mount(loading)
        message_container.scroll_end()

        # Remove the last two messages from chat history, we will regenerate them
        self.ollama.history = self.ollama.history[:-2]
        message = self.messages[-1][2]

        async def response_task() -> None:
            response = ""
            async for text in self.ollama.stream(
                message,
                [img for _, img in self.images],
                Options(seed=random.randint(0, 32768)),
            ):
                response = text
                response_chat_item.text = text
                if message_container.can_view_partial(response_chat_item):
                    message_container.scroll_end()

            # Save to db
            store = await Store.get_store()
            await store.save_message(
                id=response_message_id,
                chat_id=self.db_id,
                author=Author.OLLAMA.value,
                text=response,
            )
            self.messages.append((response_message_id, Author.OLLAMA, response, []))
            self.images = []

            loading.remove()

        asyncio.create_task(response_task())

    async def action_history(self) -> None:
        def on_history_selected(text: str | None) -> None:
            if text is None:
                return
            prompt = self.query_one("#prompt", FlexibleInput)
            if "\n" in text and not prompt.is_multiline:
                prompt.toggle_multiline()
            prompt.text = text
            prompt.focus()

        prompts = [
            message for _, author, message, _ in self.messages if author == Author.USER
        ]
        prompts.reverse()
        screen = PromptHistory(prompts)
        self.app.push_screen(screen, on_history_selected)

    @on(ImageAdded)
    def on_image_added(self, ev: ImageAdded) -> None:
        self.images.append((ev.path, ev.image))
        self.app.notify(f"Image {ev.path} added.")

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Static(f"model: {self.ollama.model}", id="info")
            yield Vertical(id="messageContainer")
            yield FlexibleInput("", id="prompt", classes="singleline")


class ChatItem(Widget):
    text: reactive[str] = reactive("")
    author: Author

    @on(Click)
    async def on_click(self, event: Click) -> None:
        try:
            pyperclip.copy(self.text)
        except pyperclip.PyperclipException:
            # https://pyperclip.readthedocs.io/en/latest/index.html#not-implemented-error
            return
        widgets = self.query(".text")
        for widget in widgets:
            widget.styles.animate("opacity", 0.5, duration=0.1)
            widget.styles.animate("opacity", 1.0, duration=0.1, delay=0.1)

    async def watch_text(self, text: str) -> None:
        text = self.text
        try:
            jsn = json.loads(text)
            if isinstance(jsn, dict):
                text = f"```json\n{self.text}\n```"
        except json.JSONDecodeError:
            pass

        txt_widget = self.query_one(".text", Markdown)
        await txt_widget.update(text)

    def compose(self) -> ComposeResult:
        """A chat item."""
        mrk_down = Markdown(self.text, classes="text")
        mrk_down.code_dark_theme = "solarized-dark"
        mrk_down.code_light_theme = "solarized-light"
        with Horizontal(classes=f"{self.author.name} chatItem"):
            yield Static(self.author.value, classes="author", markup=False)
            yield mrk_down



import json

from ollama import Options, ShowResponse
from pydantic import ValidationError
from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.containers import (
    Container,
    Horizontal,
    ScrollableContainer,
    Vertical,
)
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import Button, Checkbox, Input, Label, OptionList, TextArea

from oterm.ollamaclient import (
    OllamaLLM,
    jsonify_options,
    parse_format,
    parse_ollama_parameters,
)
from oterm.tools import available as available_tool_defs
from oterm.types import Tool


class OtermOllamaOptions(Options):
    class Config:
        extra = "forbid"


class ChatEdit(ModalScreen[str]):
    models = []
    models_info: dict[str, ShowResponse] = {}

    model_name: reactive[str] = reactive("")
    tag: reactive[str] = reactive("")
    bytes: reactive[int] = reactive(0)
    model_info: ShowResponse
    system: reactive[str] = reactive("")
    parameters: reactive[Options] = reactive(Options())
    format: reactive[str] = reactive("")
    keep_alive: reactive[int] = reactive(5)
    last_highlighted_index = None
    tools: reactive[list[Tool]] = reactive([])
    edit_mode: reactive[bool] = reactive(False)

    BINDINGS = [
        ("escape", "cancel", "Cancel"),
        ("enter", "save", "Save"),
    ]

    def __init__(
        self,
        model: str = "",
        system: str = "",
        parameters: Options = Options(),
        format="",
        keep_alive: int = 5,
        edit_mode: bool = False,
        tools: list[Tool] = [],
    ) -> None:
        super().__init__()
        self.model_name, self.tag = model.split(":") if model else ("", "")
        self.system = system
        self.parameters = parameters
        self.format = format
        self.keep_alive = keep_alive
        self.edit_mode = edit_mode
        self.tools = tools

    def _return_chat_meta(self) -> None:
        model = f"{self.model_name}:{self.tag}"
        system = self.query_one(".system", TextArea).text
        system = system if system != self.model_info.get("system", "") else None
        keep_alive = int(self.query_one(".keep-alive", Input).value)
        p_area = self.query_one(".parameters", TextArea)
        try:
            parameters = OtermOllamaOptions.model_validate_json(
                p_area.text, strict=True
            ).model_dump(exclude_unset=True)
        except ValidationError:
            p_area = self.query_one(".parameters", TextArea)
            p_area.styles.animate("opacity", 0.0, final_value=1.0, duration=0.5)
            return
        f_area = self.query_one(".format", TextArea)

        # Try parsing the format
        try:
            parse_format(f_area.text)
            format = f_area.text
        except Exception:
            f_area.styles.animate("opacity", 0.0, final_value=1.0, duration=0.5)
            return

        result = json.dumps(
            {
                "name": model,
                "system": system,
                "format": format,
                "keep_alive": keep_alive,
                "parameters": parameters,
                "tools": [tool.model_dump() for tool in self.tools],
            }
        )

        self.dismiss(result)

    def action_cancel(self) -> None:
        self.dismiss()

    def action_save(self) -> None:
        self._return_chat_meta()

    def select_model(self, model: str) -> None:
        select = self.query_one("#model-select", OptionList)
        for index, option in enumerate(select._options):
            if str(option.prompt) == model:
                select.highlighted = index
                break

    async def on_mount(self) -> None:
        self.models = OllamaLLM.list().models

        models = [model.model or "" for model in self.models]
        for model in models:
            info = OllamaLLM.show(model)
            self.models_info[model] = info
        option_list = self.query_one("#model-select", OptionList)
        option_list.clear_options()
        for model in models:
            option_list.add_option(item=self.model_option(model))
        option_list.highlighted = self.last_highlighted_index
        if self.model_name and self.tag:
            self.select_model(f"{self.model_name}:{self.tag}")

        # Disable the model select widget if we are in edit mode.
        widget = self.query_one("#model-select", OptionList)
        widget.disabled = self.edit_mode

    @on(Checkbox.Changed)
    def on_tool_toggled(self, ev: Checkbox.Changed):
        tool_name = ev.control.label
        checked = ev.value
        for tool_def in available_tool_defs:
            if tool_def["tool"].function.name == str(tool_name):  # type: ignore
                tool = tool_def["tool"]
                if checked:
                    self.tools.append(tool)
                else:
                    self.tools.remove(tool)
                break

    def on_option_list_option_highlighted(
        self, option: OptionList.OptionHighlighted
    ) -> None:
        model = option.option.prompt
        model_meta = next((m for m in self.models if m.model == str(model)), None)
        if model_meta:
            name, tag = (model_meta.model or "").split(":")
            self.model_name = name
            widget = self.query_one(".name", Label)
            widget.update(f"{self.model_name}")

            self.tag = tag
            widget = self.query_one(".tag", Label)
            widget.update(f"{self.tag}")

            self.bytes = model_meta["size"]
            widget = self.query_one(".size", Label)
            widget.update(f"{(self.bytes / 1.0e9):.2f} GB")

            meta = self.models_info.get(model_meta.model or "")
            self.model_info = meta  # type: ignore
            if not self.edit_mode:
                self.parameters = parse_ollama_parameters(
                    self.model_info.parameters or ""
                )
            widget = self.query_one(".parameters", TextArea)
            widget.load_text(jsonify_options(self.parameters))
            widget = self.query_one(".system", TextArea)

            # XXX Does not work as expected, there is no longer system in model_info
            widget.load_text(self.system or self.model_info.get("system", ""))

            # Deduce from the model's template if the model is tool-capable.
            tools_supported = ".Tools" in self.model_info["template"]
            widgets = self.query(".tool")
            for widget in widgets:
                widget.disabled = not tools_supported
                if not tools_supported:
                    widget.value = False  # type: ignore

        # Now that there is a model selected we can save the chat.
        save_button = self.query_one("#save-btn", Button)
        save_button.disabled = False
        ChatEdit.last_highlighted_index = option.option_index

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.name == "save":
            self._return_chat_meta()
        else:
            self.dismiss()

    @staticmethod
    def model_option(model: str) -> Text:
        return Text(model)

    def compose(self) -> ComposeResult:
        with Container(id="edit-chat-container"):
            with Horizontal():
                with Vertical():
                    with Horizontal(id="model-info"):
                        yield Label("Model:", classes="title")
                        yield Label(f"{self.model_name}", classes="name")
                        yield Label("Tag:", classes="title")
                        yield Label(f"{self.tag}", classes="tag")
                        yield Label("Size:", classes="title")
                        yield Label(f"{self.size}", classes="size")

                    yield OptionList(id="model-select")
                    yield Label("Tools:", classes="title")
                    with ScrollableContainer(id="tool-list"):
                        for tool_def in available_tool_defs:
                            yield Checkbox(
                                label=f"{tool_def['tool']['function']['name']}",
                                tooltip=f"{tool_def['tool']['function']['description']}",
                                value=tool_def["tool"] in self.tools,
                                classes="tool",
                            )

                with Vertical():
                    yield Label("System:", classes="title")
                    yield TextArea(self.system, classes="system log")
                    yield Label("Parameters:", classes="title")
                    yield TextArea(
                        jsonify_options(self.parameters),
                        classes="parameters log",
                        language="json",
                    )
                    yield Label("Format:", classes="title")
                    yield TextArea(
                        self.format or "",
                        classes="format log",
                        language="json",
                    )

                    with Horizontal():
                        with Horizontal():
                            yield Label(
                                "Keep-alive (min)", classes="title keep-alive-label"
                            )
                            yield Input(classes="keep-alive", value="5")

            with Horizontal(classes="button-container"):
                yield Button(
                    "Save",
                    id="save-btn",
                    name="save",
                    disabled=True,
                    variant="primary",
                )
                yield Button("Cancel", name="cancel")



import pytest
from ollama import ResponseError

from oterm.ollamaclient import OllamaLLM
from oterm.tools.location import LocationTool


@pytest.mark.asyncio
async def test_generate():
    llm = OllamaLLM()
    res = await llm.completion(prompt="Please add 42 and 42")
    assert "84" in res


@pytest.mark.asyncio
async def test_llm_context():
    llm = OllamaLLM()
    await llm.completion("I am testing oterm, a python client for Ollama.")
    # There should now be a context saved for the conversation.
    res = await llm.completion("Do you remember what I am testing?")
    assert "oterm" in res.lower()


@pytest.mark.asyncio
async def test_multi_modal_llm(llama_image):
    llm = OllamaLLM(model="llava")
    res = await llm.completion("Describe this image", images=[llama_image])
    assert "llama" in res or "animal" in res


@pytest.mark.asyncio
async def test_errors():
    llm = OllamaLLM(model="non-existent-model")
    try:
        await llm.completion("This should fail.")
    except ResponseError as e:
        assert 'model "non-existent-model" not found' in str(e)


@pytest.mark.asyncio
async def test_iterator():
    llm = OllamaLLM()
    response = ""
    async for text in llm.stream("Please add 2 and 2"):
        response = text
    assert "4" in response


@pytest.mark.skip(
    reason="Skipped till https://github.com/ollama/ollama-python/issues/279 is fixed."
)
@pytest.mark.asyncio
async def test_tool_streaming():
    # This test will fail until Ollama supports streaming with tools.
    # See https://github.com/ollama/ollama-python/issues/279

    llm = OllamaLLM(
        tool_defs=[
            {"tool": LocationTool, "callable": lambda: "New York"},
        ],
    )
    response = ""
    async for text in llm.stream(
        "In which city am I currently located?. Reply with no other text, just the city."
    ):
        response = text
    assert "New York" in response



import json

import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.location import LocationTool, current_location
from oterm.tools.weather import WeatherTool, current_weather


@pytest.mark.asyncio
async def test_weather():
    llm = OllamaLLM(
        tool_defs=[
            {"tool": WeatherTool, "callable": current_weather},
        ],
    )
    weather = json.loads(await current_weather(latitude=59.2675, longitude=10.4076))
    temperature = weather.get("main").get("temp") - 273.15

    res = await llm.completion(
        "What is the current temperature at my location latitude 59.2675, longitude 10.4076?"
    )

    assert "temperature" in res or "Temperature" in res
    assert str(round(temperature)) in res or str(round(temperature, 1)) in res


@pytest.mark.asyncio
async def test_weather_with_location():
    llm = OllamaLLM(
        tool_defs=[
            {"tool": LocationTool, "callable": current_location},
            {"tool": WeatherTool, "callable": current_weather},
        ],
    )
    location = json.loads(await current_location())
    weather = json.loads(
        await current_weather(
            latitude=location.get("latitude"),
            longitude=location.get("longitude"),
        )
    )
    temperature = weather.get("main").get("temp") - 273.15

    res = await llm.completion("What is the current temperature in my city?")
    assert "temperature" in res or "Temperature" in res
    assert str(round(temperature)) in res or str(round(temperature, 1)) in res



from enum import Enum
from typing import Awaitable, Callable, TypedDict

from ollama._types import Image, Tool  # noqa


class Author(Enum):
    USER = "me"
    OLLAMA = "ollama"


class ParsedResponse(TypedDict):
    thought: str
    response: str
    formatted_output: str


class ToolDefinition(TypedDict):
    tool: Tool
    callable: Callable | Awaitable


class ExternalToolDefinition(TypedDict):
    tool: str
    callable: str



from io import BytesIO

import ollama
import pytest
import pytest_asyncio
from PIL import Image


@pytest_asyncio.fixture(autouse=True)
async def load_test_models():
    try:
        ollama.show("llama3.2")
    except ollama.ResponseError:
        ollama.pull("llama3.2")
    yield


@pytest.fixture(scope="session")
def llama_image() -> bytes:
    buffered = BytesIO()
    image = Image.open("tests/data/lama.jpg")
    image.save(buffered, format="JPEG")
    return buffered.getvalue()



from mcp import ClientSession, ListToolsResult, StdioServerParameters
from mcp import Tool as MCPTool
from mcp.client.stdio import stdio_client
from mcp.types import CallToolResult

from oterm.types import Tool


class MCPClient:
    """Client for interacting with MCP servers"""

    def __init__(self, server_params: StdioServerParameters):
        self.server_params = server_params
        self.session = None
        self._client = None

    async def __aenter__(self):
        """Async context manager entry"""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit"""
        if self.session:
            await self.session.__aexit__(exc_type, exc_val, exc_tb)
        if self._client:
            await self._client.__aexit__(exc_type, exc_val, exc_tb)

    async def connect(self):
        self._client = stdio_client(self.server_params)
        self.read, self.write = await self._client.__aenter__()
        session = ClientSession(self.read, self.write)
        self.session = await session.__aenter__()
        await self.session.initialize()

    async def get_available_tools(self) -> list[MCPTool]:
        """List available tools"""
        if not self.session:
            raise RuntimeError("Not connected to MCP server")
        tools: ListToolsResult = await self.session.list_tools()
        # Let's just ignore pagination for now
        return tools.tools

    async def call_tool(self, tool_name: str, arguments: dict) -> CallToolResult:
        """Call a tool with given arguments"""
        if not self.session:
            raise RuntimeError("Not connected to MCP server")

        result = await self.session.call_tool(tool_name, arguments=arguments)
        return result


def mcp_tool_to_ollama_tool(mcp_tool: MCPTool) -> Tool:
    """Convert an MCP tool to an Ollama tool"""

    return Tool(
        function=Tool.Function(
            name=mcp_tool.name,
            description=mcp_tool.description,
            parameters=Tool.Function.Parameters.model_validate(mcp_tool.inputSchema),
        ),
    )



import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.web import WebTool, fetch_url


@pytest.mark.asyncio
async def test_web():
    llm = OllamaLLM(
        tool_defs=[{"tool": WebTool, "callable": fetch_url}],
    )
    res = await llm.completion(
        "What's oterm in a single phrase? oterm is hosted at https://github.com/ggozad/oterm."
    )
    assert "Ollama" in res



import inspect
import json
from ast import literal_eval
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Iterator,
    Literal,
    Mapping,
    Sequence,
)

from ollama import (
    AsyncClient,
    ChatResponse,
    Client,
    ListResponse,
    Message,
    Options,
    ProgressResponse,
    ShowResponse,
)
from pydantic.json_schema import JsonSchemaValue
from textual import log

from oterm.config import envConfig
from oterm.types import ToolDefinition


def parse_format(format_text: str) -> JsonSchemaValue | Literal["", "json"]:
    try:
        jsn = json.loads(format_text)
        if isinstance(jsn, dict):
            return jsn
    except json.JSONDecodeError:
        if format_text in ("", "json"):
            return format_text
    raise Exception(f"Invalid Ollama format: '{format_text}'")


class OllamaLLM:
    def __init__(
        self,
        model="llama3.2",
        system: str | None = None,
        history: list[Mapping[str, Any] | Message] = [],
        format: str = "",
        options: Options = Options(),
        keep_alive: int = 5,
        tool_defs: Sequence[ToolDefinition] = [],
    ):
        self.model = model
        self.system = system
        self.history = history
        self.format = format
        self.keep_alive = keep_alive
        self.options = options
        self.tool_defs = tool_defs
        self.tools = [tool["tool"] for tool in tool_defs]

        if system:
            system_prompt: Message = Message(role="system", content=system)
            self.history = [system_prompt] + self.history

    async def completion(
        self,
        prompt: str = "",
        images: list[Path | bytes | str] = [],
        tool_call_messages=[],
    ) -> str:
        client = AsyncClient(
            host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL
        )
        if prompt:
            user_prompt: Message = Message(role="user", content=prompt)
            if images:
                # This is a bug in Ollama the images should be a list of Image objects
                # user_prompt.images = [Image(value=image) for image in images]
                user_prompt.images = images  # type: ignore
            self.history.append(user_prompt)
        response: ChatResponse = await client.chat(
            model=self.model,
            messages=self.history + tool_call_messages,
            keep_alive=f"{self.keep_alive}m",
            options=self.options,
            format=parse_format(self.format),
            tools=self.tools,
        )
        message = response.message
        tool_calls = message.tool_calls
        if tool_calls:
            tool_messages = [message]
            for tool_call in tool_calls:

                tool_name = tool_call["function"]["name"]
                for tool_def in self.tool_defs:
                    log.debug("Calling tool: %s", tool_name)
                    if tool_def["tool"]["function"]["name"] == tool_name:
                        tool_callable = tool_def["callable"]
                        tool_arguments = tool_call["function"]["arguments"]
                        try:
                            if inspect.iscoroutinefunction(tool_callable):
                                tool_response = await tool_callable(**tool_arguments)  # type: ignore
                            else:
                                tool_response = tool_callable(**tool_arguments)  # type: ignore
                            log.debug(f"Tool response: {tool_response}", tool_response)
                        except Exception as e:
                            log.error(f"Error calling tool {tool_name}", e)
                            tool_response = str(e)
                        tool_messages.append(
                            {  # type: ignore
                                "role": "tool",
                                "content": tool_response,
                                "name": tool_name,
                            }
                        )
            return await self.completion(
                tool_call_messages=tool_messages,
            )

        self.history.append(message)
        text_response = message.content
        return text_response or ""

    async def stream(
        self,
        prompt: str,
        images: list[Path | bytes | str] = [],
        additional_options: Options = Options(),
        tool_defs: Sequence[ToolDefinition] = [],
    ) -> AsyncGenerator[str, Any]:

        # stream() should not be called with tools till Ollama supports streaming with tools.
        # See https://github.com/ollama/ollama-python/issues/279
        if tool_defs:
            raise NotImplementedError(
                "stream() should not be called with tools till Ollama supports streaming with tools."
            )

        client = AsyncClient(
            host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL
        )
        user_prompt: Message = Message(role="user", content=prompt)
        if images:
            user_prompt.images = images  # type: ignore

        self.history.append(user_prompt)
        options = {
            k: v for k, v in self.options.model_dump().items() if v is not None
        } | {k: v for k, v in additional_options.model_dump().items() if v is not None}

        stream: AsyncIterator[ChatResponse] = await client.chat(
            model=self.model,
            messages=self.history,
            stream=True,
            options=options,
            keep_alive=f"{self.keep_alive}m",
            format=parse_format(self.format),
            tools=self.tools,
        )
        text = ""
        async for response in stream:
            text = text + response.message.content if response.message.content else text
            yield text

        self.history.append(Message(role="assistant", content=text))

    @staticmethod
    def list() -> ListResponse:
        client = Client(host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL)
        return client.list()

    @staticmethod
    def show(model: str) -> ShowResponse:
        client = Client(host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL)
        return client.show(model)

    @staticmethod
    def pull(model: str) -> Iterator[ProgressResponse]:
        client = Client(host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL)
        stream: Iterator[ProgressResponse] = client.pull(model, stream=True)
        for response in stream:
            yield response


def parse_ollama_parameters(parameter_text: str) -> Options:
    lines = parameter_text.split("\n")
    params = Options()
    valid_params = set(Options.model_fields.keys())
    for line in lines:
        if line:
            key, value = line.split(maxsplit=1)
            try:
                value = literal_eval(value)
            except (SyntaxError, ValueError):
                pass
            if key not in valid_params:
                continue
            if params.get(key):
                if not isinstance(params[key], list):
                    params[key] = [params[key], value]
                else:
                    params[key].append(value)
            else:
                params[key] = value
    return params


def jsonify_options(options: Options) -> str:
    return json.dumps(
        {
            key: value
            for key, value in options.model_dump().items()
            if value is not None
        },
        indent=2,
    )



import json
from html.parser import HTMLParser

import httpx

from oterm.types import Tool

WebTool = Tool(
    type="function",
    function=Tool.Function(
        name="fetch_url",
        description="Function to return the contents of a website in text format.",
        parameters=Tool.Function.Parameters(
            type="object",
            properties={
                "url": Tool.Function.Parameters.Property(
                    type="str", description="The URL of the website to fetch."
                ),
            },
            required=["url"],
        ),
    ),
)


class HTML2Text(HTMLParser):
    text = ""

    def handle_data(self, data):
        self.text += data


async def fetch_url(url: str) -> str:
    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(url)

            if response.status_code == 200:
                html = response.text
                parser = HTML2Text()
                parser.feed(html)
                return parser.text

            else:
                return json.dumps(
                    {
                        "error": f"Failed to fetch URL: {url}. Status code: {response.status_code}"
                    }
                )
        except Exception as e:
            return json.dumps({"error": str(e)})



from importlib import import_module
from typing import Awaitable, Callable, Sequence

from mcp import StdioServerParameters
from mcp import Tool as MCPTool
from mcp.types import CallToolResult, TextContent
from ollama._types import Tool

from oterm.config import appConfig
from oterm.tools.date_time import DateTimeTool, date_time
from oterm.tools.location import LocationTool, current_location
from oterm.tools.mcp import MCPClient, mcp_tool_to_ollama_tool
from oterm.tools.shell import ShellTool, shell_command
from oterm.tools.weather import WeatherTool, current_weather
from oterm.tools.web import WebTool, fetch_url
from oterm.types import ExternalToolDefinition, ToolDefinition


def load_tools(tool_defs: Sequence[ExternalToolDefinition]) -> Sequence[ToolDefinition]:
    tools = []
    for tool_def in tool_defs:
        tool_path = tool_def["tool"]

        try:
            module, tool = tool_path.split(":")
            module = import_module(module)
            tool = getattr(module, tool)
            if not isinstance(tool, Tool):
                raise Exception(f"Expected Tool, got {type(tool)}")
        except ModuleNotFoundError as e:
            raise Exception(f"Error loading tool {tool_path}: {str(e)}")

        callable_path = tool_def["callable"]
        try:
            module, function = callable_path.split(":")
            module = import_module(module)
            callable = getattr(module, function)
            if not isinstance(callable, (Callable, Awaitable)):
                raise Exception(f"Expected Callable, got {type(callable)}")
        except ModuleNotFoundError as e:
            raise Exception(f"Error loading callable {callable_path}: {str(e)}")
        tools.append({"tool": tool, "callable": callable})

    return tools


available: list[ToolDefinition] = []

external_tools = appConfig.get("tools")
if external_tools:
    available.extend(load_tools(external_tools))


async def setup_mcp_servers():
    mcp_servers = appConfig.get("mcpServers")
    tool_defs: list[ToolDefinition] = []
    if mcp_servers:
        for server, config in mcp_servers.items():
            async with MCPClient(
                StdioServerParameters.model_validate(config)
            ) as client:
                mcp_tools: list[MCPTool] = await client.get_available_tools()

                for mcp_tool in mcp_tools:
                    tool = mcp_tool_to_ollama_tool(mcp_tool)

                    async def callable(name=mcp_tool.name, **kwargs):
                        async with MCPClient(
                            StdioServerParameters.model_validate(config)
                        ) as cl:
                            res: CallToolResult = await cl.call_tool(name, kwargs)
                            if res.isError:
                                raise Exception(f"Error call mcp tool {name}.")
                            text_content = [
                                m.text for m in res.content if type(m) is TextContent
                            ]
                            return "\n".join(text_content)

                    tool_defs.append({"tool": tool, "callable": callable})

    return tool_defs



import json

import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.location import LocationTool, current_location


@pytest.mark.asyncio
async def test_location_tool():
    llm = OllamaLLM(
        model="mistral-nemo",
        tool_defs=[
            {"tool": LocationTool, "callable": current_location},
        ],
    )
    res = await llm.completion(
        "In which city am I currently located?. Reply with no other text, just the city."
    )
    curr_loc = json.loads(await current_location()).get("city")
    assert curr_loc in res



import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.shell import ShellTool, shell_command


@pytest.mark.asyncio
async def test_shell():
    llm = OllamaLLM(
        model="mistral-nemo",
        tool_defs=[
            {"tool": ShellTool, "callable": shell_command},
        ],
    )
    res = await llm.completion("What is the current directory")
    assert "oterm" in res



from datetime import datetime

import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.date_time import DateTimeTool, date_time


@pytest.mark.asyncio
async def test_date_time():
    llm = OllamaLLM(
        model="mistral-nemo", tool_defs=[{"tool": DateTimeTool, "callable": date_time}]
    )
    res = await llm.completion(
        "What is the current date in YYYY-MM-DD format?. Reply with no other text, just the date."
    )
    date = datetime.date(datetime.now())
    assert f"{date.year}-{date.month:02d}-{date.day:02d}" in res



import json
from typing import Iterable

from ollama import Options, Tool
from textual import on, work
from textual.app import App, ComposeResult, SystemCommand
from textual.binding import Binding
from textual.screen import Screen
from textual.widgets import Footer, Header, TabbedContent, TabPane

from oterm.app.chat_edit import ChatEdit
from oterm.app.chat_export import ChatExport, slugify
from oterm.app.pull_model import PullModel
from oterm.app.splash import splash
from oterm.app.widgets.chat import ChatContainer
from oterm.config import appConfig
from oterm.store.store import Store
from oterm.tools import setup_mcp_servers


class OTerm(App):
    TITLE = "oTerm"
    SUB_TITLE = "A terminal-based Ollama client."
    CSS_PATH = "oterm.tcss"
    BINDINGS = [
        Binding("ctrl+tab", "cycle_chat(+1)", "next chat", id="next.chat"),
        Binding("ctrl+shift+tab", "cycle_chat(-1)", "prev chat", id="prev.chat"),
        Binding("ctrl+backspace", "delete_chat", "delete chat", id="delete.chat"),
        Binding("ctrl+n", "new_chat", "new chat", id="new.chat"),
        Binding("ctrl+q", "quit", "quit", id="quit"),
    ]

    def get_system_commands(self, screen: Screen) -> Iterable[SystemCommand]:
        yield from super().get_system_commands(screen)
        yield SystemCommand("New chat", "Creates a new chat", self.action_new_chat)
        yield SystemCommand(
            "Edit chat parameters",
            "Allows to redefine model parameters and system prompt",
            self.action_edit_chat,
        )
        yield SystemCommand(
            "Rename chat", "Renames the current chat", self.action_rename_chat
        )
        yield SystemCommand(
            "Clear chat", "Clears the current chat", self.action_clear_chat
        )
        yield SystemCommand(
            "Delete chat", "Deletes the current chat", self.action_delete_chat
        )
        yield SystemCommand(
            "Export chat",
            "Exports the current chat as Markdown (in the current working directory)",
            self.action_export_chat,
        )
        yield SystemCommand(
            "Regenerate last Ollama message",
            "Regenerates the last Ollama message (setting a random seed for the message)",
            self.action_regenerate_last_message,
        )

        yield SystemCommand(
            "Pull model",
            "Pulls (or updates) the model from the Ollama server",
            self.action_pull_model,
        )

    async def action_quit(self) -> None:
        return self.exit()

    async def action_cycle_chat(self, change: int) -> None:
        tabs = self.query_one(TabbedContent)
        store = await Store.get_store()
        saved_chats = await store.get_chats()
        if tabs.active_pane is None:
            return
        active_id = int(str(tabs.active_pane.id).split("-")[1])
        for _chat in saved_chats:
            if _chat[0] == active_id:
                next_index = (saved_chats.index(_chat) + change) % len(saved_chats)
                next_id = saved_chats[next_index][0]
                tabs.active = f"chat-{next_id}"
                break

    @work
    async def action_new_chat(self) -> None:
        store = await Store.get_store()
        model_info: str | None = await self.push_screen_wait(ChatEdit())
        if not model_info:
            return
        model: dict = json.loads(model_info)
        tabs = self.query_one(TabbedContent)
        tab_count = tabs.tab_count
        name = f"chat #{tab_count+1} - {model['name']}"
        id = await store.save_chat(
            id=None,
            name=name,
            model=model["name"],
            system=model["system"],
            format=model["format"],
            parameters=model["parameters"],
            keep_alive=model["keep_alive"],
            tools=model["tools"],
        )
        pane = TabPane(name, id=f"chat-{id}")
        pane.compose_add_child(
            ChatContainer(
                db_id=id,
                chat_name=name,
                model=model["name"],
                system=model["system"],
                format=model["format"],
                parameters=Options(**model.get("parameters", {})),
                keep_alive=model["keep_alive"],
                messages=[],
                tools=[Tool(**t) for t in model.get("tools", [])],
            )
        )
        await tabs.add_pane(pane)
        tabs.active = f"chat-{id}"

    async def action_edit_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        chat.action_edit_chat()

    async def action_rename_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        chat.action_rename_chat()

    async def action_clear_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        await chat.action_clear_chat()

    async def action_delete_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        store = await Store.get_store()
        await store.delete_chat(chat.db_id)
        await tabs.remove_pane(tabs.active)

    async def action_export_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        screen = ChatExport()
        screen.chat_id = chat.db_id
        screen.file_name = f"{slugify(chat.chat_name)}.md"
        self.push_screen(screen)

    async def action_regenerate_last_message(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        await chat.action_regenerate_llm_message()

    async def action_pull_model(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            screen = PullModel("")
        else:
            chat = tabs.active_pane.query_one(ChatContainer)
            screen = PullModel(chat.ollama.model)
        self.push_screen(screen)

    async def load_mcp(self):
        from oterm.tools import available

        mcp_tool_defs = await setup_mcp_servers()
        available += mcp_tool_defs

    async def on_mount(self) -> None:
        store = await Store.get_store()
        theme = appConfig.get("theme")
        if theme:
            if theme == "dark":
                self.theme = "textual-dark"
            elif theme == "light":
                self.theme = "textual-light"
            else:
                self.theme = theme
        self.dark = appConfig.get("theme") == "dark"
        self.watch(self.app, "theme", self.on_theme_change, init=False)

        saved_chats = await store.get_chats()
        # Apply any remap of key bindings.
        keymap = appConfig.get("keymap")
        if keymap:
            self.set_keymap(keymap)

        await self.load_mcp()

        async def on_splash_done(message) -> None:
            if not saved_chats:
                # Pyright suggests awaiting here which has bitten me twice
                # so I'm ignoring it
                self.action_new_chat()  # type: ignore
            else:
                tabs = self.query_one(TabbedContent)
                for (
                    id,
                    name,
                    model,
                    system,
                    format,
                    parameters,
                    keep_alive,
                    tools,
                ) in saved_chats:
                    messages = await store.get_messages(id)
                    container = ChatContainer(
                        db_id=id,
                        chat_name=name,
                        model=model,
                        messages=messages,
                        system=system,
                        format=format,
                        parameters=parameters,
                        keep_alive=keep_alive,
                        tools=tools,
                    )
                    pane = TabPane(name, container, id=f"chat-{id}")
                    tabs.add_pane(pane)

        if appConfig.get("splash-screen"):
            self.push_screen(splash, callback=on_splash_done)
        else:
            await on_splash_done("")

    def on_theme_change(self, old_value: str, new_value: str) -> None:
        if appConfig.get("theme") != new_value:
            appConfig.set("theme", new_value)

    @work
    @on(TabbedContent.TabActivated)
    async def on_tab_activated(self, event: TabbedContent.TabActivated) -> None:
        container = event.pane.query_one(ChatContainer)
        await container.load_messages()

    def compose(self) -> ComposeResult:
        yield Header()
        yield TabbedContent(id="tabs")
        yield Footer()


app = OTerm()



import json

import httpx

from oterm.config import envConfig
from oterm.types import Tool

WeatherTool = Tool(
    type="function",
    function=Tool.Function(
        name="current_weather",
        description="Function to return the current weather for the given location in Standard Units.",
        parameters=Tool.Function.Parameters(
            type="object",
            properties={
                "latitude": Tool.Function.Parameters.Property(
                    type="float", description="The latitude of the location."
                ),
                "longitude": Tool.Function.Parameters.Property(
                    type="float", description="The longitude of the location."
                ),
            },
            required=["latitude", "longitude"],
        ),
    ),
)


async def current_weather(latitude: float, longitude: float) -> str:
    async with httpx.AsyncClient() as client:
        try:
            api_key = envConfig.OPEN_WEATHER_MAP_API_KEY
            if not api_key:
                raise Exception("OpenWeatherMap API key not found")

            response = await client.get(
                f"https://api.openweathermap.org/data/2.5/weather?lat={latitude}&lon={longitude}&appid={api_key}"
            )

            if response.status_code == 200:
                data = response.json()
                return json.dumps(data)
            else:
                return json.dumps(
                    {"error": f"{response.status_code}: {response.reason_phrase}"}
                )
        except Exception as e:
            return json.dumps({"error": str(e)})



import json

import httpx

from oterm.types import Tool

LocationTool = Tool(
    type="function",
    function=Tool.Function(
        name="current_location",
        description="Function to return the current location, city, region, country, latitude, and longitude.",
        parameters=Tool.Function.Parameters(
            type="object",
            properties={},
            required=[],
        ),
    ),
)


async def current_location():

    async with httpx.AsyncClient() as client:
        try:
            response = await client.get("https://ipinfo.io/")
            if response.status_code == 200:
                data = response.json()

                # Extract latitude and longitude from the location information
                city = data.get("city", "N/A")
                region = data.get("region", "N/A")
                country = data.get("country", "N/A")
                loc = data.get("loc", "N/A").split(",")
                if len(loc) == 2:
                    latitude, longitude = float(loc[0]), float(loc[1])
                else:
                    latitude, longitude = None, None

                return json.dumps(
                    {
                        "city": city,
                        "region": region,
                        "country": country,
                        "latitude": latitude,
                        "longitude": longitude,
                    }
                )
            else:
                return json.dumps(
                    {"error": f"{response.status_code}: {response.reason_phrase}"}
                )
        except httpx.HTTPError as e:
            return json.dumps({"error": str(e)})
